<'
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "<"$comp">/c/client.h"
#include "<"$comp">_portlib.h"
#include "<"$comp">_msglib.h"
#include "private.h"


<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} continue'>

/* <"[--- ${comp}_[$s name]_action_cb -----------------------------------]"> */

void
<"$comp">_<"[$s name]">_action_cb(
  genom_client h, int rqstid, gnmix_<"[$s name]">::GoalHandle g)
{
  if (g.getCommState() == actionlib::CommState::ACTIVE) {
    pthread_mutex_lock(&h->activelk);
    if (h->active_rqst[rqstid].status != ROS_CLIENT_WAIT_TERM) {
      h->active_rqst[rqstid].status = ROS_CLIENT_WAIT_TERM;
      h->events = 1;
    }
    pthread_mutex_unlock(&h->activelk);

  } else if (g.getCommState() == actionlib::CommState::DONE) {
    const boost::shared_ptr<const genom::srv::<"[$s name]">::output> rout =
      g.getResult();

    pthread_mutex_lock(&h->activelk);
    if (rout) {
      if (rout->genom_exception) {
        h->active_rqst[rqstid].error =
          genom_<"$comp">_decodex(
            rout->genom_exdetail.c_str(), h->active_rqst[rqstid].exdetail,
            h->active_rqst[rqstid].exfini);
      } else
        h->active_rqst[rqstid].error = genom_ok;

<'  foreach p [$s parameter inout out] {'>
      genom::ids::pcopy(
        ((struct genom_<"$comp">_<"[$s name]">_output *)
         h->active_rqst[rqstid].output)-><"[$p name]">,
        rout-><"[$p name]">);
<'  }'>
    } else {
      genom_syserr_detail *d = (genom_syserr_detail *)malloc(sizeof *d);
      if (d) d->code = ENXIO;
      h->active_rqst[rqstid].exdetail = (void *)d;
      h->active_rqst[rqstid].error = genom_syserr_id;
    }
    h->active_rqst[rqstid].status = ROS_CLIENT_TERM;
    h->active_rqst[rqstid].gh.reset();
    h->events = 1;
    pthread_mutex_unlock(&h->activelk);
  }
}

<'}'>

/* <"[--- ${comp}_abort -------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_abort(genom_client h, int rqstid)
{
  genom_gh gh;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= (signed)h->active_rqst.size() ||
      h->active_rqst[rqstid].status == ROS_CLIENT_FREE) {
    pthread_mutex_unlock(&h->activelk);
    return genom_no_such_activity();
  }
  if (h->active_rqst[rqstid].status == ROS_CLIENT_TERM) {
    pthread_mutex_unlock(&h->activelk);
    return genom_ok;
  }
  gh = h->active_rqst[rqstid].gh;
  pthread_mutex_unlock(&h->activelk);

  gh.cancel();
  return genom_ok;
}

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set svc [catch {$s task}]'>
  genom::srv::<"[$s name]">::input rin;
  assert(h);

<'  if {!$svc} {'>
  if (!h->connections-><"[$s name]">->isServerConnected()) {
    genom_syserr_detail d = { ENXIO };
    return genom_syserr(&d);
  }
<'  }'>

<'  foreach p [$s parameter in inout] {'>
  genom::ids::pcopy(rin.<"[$p name]">, in-><"[$p name]">);
<'  }'>

  h->error = ros_newrqst(
    h, sizeof(struct genom_<"$comp">_<"[$s name]">_output), *rqstid);
  if (h->error) return h->error;

  genom_<"$comp">_client_<"[$s name]">_init_output(
    (struct genom_<"$comp">_<"[$s name]">_output *)
    h->active_rqst[*rqstid].output);
  h->active_rqst[*rqstid].info =
    genom_<"$comp">_client_<"[$s name]">_info;
  h->active_rqst[*rqstid].sentcb = (genom_request_cb)sentcb;
  h->active_rqst[*rqstid].donecb = (genom_request_cb)donecb;
  h->active_rqst[*rqstid].cb_data = cb_data;

<'  if {$svc} {'>
  h->active_rqst[*rqstid].status = ROS_CLIENT_WAIT_TERM;
  genom::srv::<"[$s name]">::output rout;
  bool r = h->connections-><"[$s name]">.call(rin, rout);

  pthread_mutex_lock(&h->activelk);
  h->active_rqst[*rqstid].status = ROS_CLIENT_TERM;
  if (!r) {
    pthread_mutex_unlock(&h->activelk);
    ros_clean(h, *rqstid);

    genom_syserr_detail d = { ENXIO };
    return genom_syserr(&d);
  } else {
<'    foreach p [$s parameter inout out] {'>
    genom::ids::pcopy(
      ((struct genom_<"$comp">_<"[$s name]">_output *)
       h->active_rqst[*rqstid].output)-><"[$p name]">,
      rout.<"[$p name]">);
<'    }'>
    if (rout.genom_exception) {
      h->active_rqst[*rqstid].error =
        genom_<"$comp">_decodex(
          rout.genom_exdetail.c_str(), h->active_rqst[*rqstid].exdetail,
          h->active_rqst[*rqstid].exfini);
    }
  }
  h->events = 1;
  pthread_mutex_unlock(&h->activelk);
  ros_evtask_signal(); /* signal the callback thread */
<'  } else {'>

  h->active_rqst[*rqstid].status = ROS_CLIENT_WAIT_ACK;
  h->active_rqst[*rqstid].gh =
    h->connections-><"[$s name]">->sendGoal(
      rin, boost::bind(&<"$comp">_<"[$s name]">_action_cb, h, *rqstid, _1),
      NULL);
<'  }'>

  return genom_ok;
}


/* <"[--- ${comp}_[$s name]_result --------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_result(
  genom_client h, int rqstid, genom_event *report,
  struct genom_<"$comp">_<"[$s name]">_output **out, void **exdetail)
{
  genom_client_s::request *a;
  genom_event s;
  void *p;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= (signed)h->active_rqst.size()) {
    pthread_mutex_unlock(&h->activelk);
    genom_syserr_detail d = { ENOENT };
    return genom_syserr(&d);
  }

  a = &h->active_rqst[rqstid];
  if (a->info != genom_<"$comp">_client_<"[$s name]">_info) {
    pthread_mutex_unlock(&h->activelk);
    genom_syserr_detail d = { ENOMSG };
    return genom_syserr(&d);
  }
  pthread_mutex_unlock(&h->activelk);

  s = ros_result(h, rqstid, report, &p, exdetail);
  if (s) return s;
  if (out) *out = (struct genom_<"$comp">_<"[$s name]">_output *)p;
  return genom_ok;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- ${comp}_[$p name]_port_cb -------------------------------------]"> */

void
<"$comp">_<"[$p name]">_port_cb(
  int &done,
  <"[[$p datatype] argument reference out]">,
  const boost::shared_ptr<genom::port::<"[$p name]"> const> &data)
{
  genom::ids::pcopy(<"[[$p datatype] dereference reference out]">, data->v);
  done = 1;
}
<'}'>

<'foreach p [$component ports out simple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  <"[[$p datatype] argument reference data]">)
{
  ros::Subscriber s;
  int delay, done = 0;

  s = h->node->subscribe< genom::port::<"[$p name]"> >(
    "/" + std::string(h->instance) + "/<"[$p name]">", 1,
    boost::bind(&<"$comp">_<"[$p name]">_port_cb, boost::ref(done), data, _1));
  if (!s) {
    genom_syserr_detail d = { ENOENT };
    return h->error = genom_syserr(&d);
  }

  delay = 0;
  while(!done) {
    struct timespec req = {0, 1000000/*1ms*/}, rem;
    delay ++;
    if (delay > 3000/*3s*/) {
      genom_syserr_detail d = { ENOENT };
      return h->error = genom_syserr(&d);
    }
    while (nanosleep(&req, &rem) == EINTR) req = rem;
  }

  return genom_ok;
}

<'}'>
<'foreach p [$component ports out multiple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  const char *path,
  <"[[$p datatype] argument reference data]">)
{
  ros::Subscriber s;
  int delay, done = 0;

  s = h->node->subscribe< genom::port::<"[$p name]"> >(
    "/" + std::string(h->instance) + "/<"[$p name]">/" + path, 1,
    boost::bind(&<"$comp">_<"[$p name]">_port_cb, boost::ref(done), data, _1));
  if (!s) {
    genom_syserr_detail d = { ENOENT };
    return h->error = genom_syserr(&d);
  }

  delay = 0;
  while(!done) {
    struct timespec req = {0, 1000000/*1ms*/}, rem;
    delay ++;
    if (delay > 3000/*3s*/) {
      genom_syserr_detail d = { ENOENT };
      return h->error = genom_syserr(&d);
    }
    while (nanosleep(&req, &rem) == EINTR) req = rem;
  }

  return genom_ok;
}

<'}'>
